#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
from cffi import FFI

if sys.version_info[0] < 3:
    # Python 2 compatibility; requires `future` package.
    from builtins import bytes

from . import config
from .dio import QmixDigitalIO
from .tools import CHK, find_dll
from .headers import VALVE_HEADER


class QmixValve(object):
    """
    Qmix valve interface.

    Parameters
    ----------
    index : int, or None
        Index of the valve to initialize. Takes precedence over the
        `name` and `handle` parameters.

    name : str
        The name of the valve to initialize. Will be ignored if `index` is
        not `None`. Takes precedence over the `handle` parameter.

    handle :  Qmix valve handle, or None
        A Qmix valve device handle, as returned by
        :func:~`pyqmix.QmixPump.valve_handle`.

    """
    def __init__(self, index=None, name='', handle=None):
        if index is None and name == '' and handle is None:
            raise ValueError('Please specify a valid valve index or name.')
        else:
            self.index = index
            self.name = name
            self.handle = handle

        dll_dir = config.read_config().get('qmix_dll_dir', None)
        dll_filename = 'labbCAN_Valve_API.dll'

        dll_path = find_dll(dll_dir=dll_dir, dll_filename=dll_filename)
        if dll_path is None:
            msg = 'Could not find the Qmix SDK DLL %s.' % dll_filename
            raise RuntimeError(msg)
        else:
            self.dll_path = dll_path

        self._ffi = FFI()
        self._ffi.cdef(VALVE_HEADER)
        self._dll = self._ffi.dlopen(self.dll_path)

        if self.index is not None:
            self._handle = self._ffi.new('dev_hdl *', 0)
            self._call('LCV_GetValveHandle', self.index, self._handle)
        elif self.name != '':
            self._call('LCV_LookupValveByName',
                       bytes(self.name, 'utf8'),
                       self._handle)
        else:
            self._handle = handle

        self.aspirate_pos = 1
        self.dispense_pos = 0

        self.name = name

    def _call(self, func_name, *args):
        func = getattr(self._dll, func_name)
        r = func(*args)
        return CHK(r)

    @property
    def number_of_positions(self):
        """
        Return the number of valve positions.

        Each valve has a number of available valve positions.
        A switching valve has two positions.

        Returns
        -------
        int
           >0 Number of valve positions

        """
        return self._call('LCV_NumberOfValvePositions', self._handle[0])

    @property
    def position(self):
        """
        The current logical valve position.

        Each valve position is identified by a logical valve position identifier
        from 0 to number of valve positions - 1. This function returns the
        logical valve position identifier for the current valve position.

        Returns
        -------
        int
           >=0 current valve position index.

        """
        return self._call('LCV_ActualValvePosition', self._handle[0])

    @position.setter
    def position(self, position):
        self.switch_position(position)

    def switch_position(self, position=None):
        """
        Switch the valve to a certain logical valve position.

        Parameters
        ----------
        position : int, or None
            Switch the valve to the specified position.
            If `None` and a valve with two possible positions is connected,
            switch from the current position to the other one.

        """
        if position is None and self.number_of_positions != 2:
            msg = ('For valves with more than 2 positions, please specify the '
                   'desired target position.')
            raise ValueError(msg)

        if position is None:
            if self.position == 0:
                target_position = 1
            else:
                target_position = 0
        else:
            target_position = position

        if ((target_position < 0) or
                (target_position >= self.number_of_positions)):
            msg = ('Must specify position in the range [0, %i] for this valve.'
                   % (self.number_of_positions - 1))
            raise ValueError(msg)

        self._call('LCV_SwitchValveToPosition', self._handle[0],
                   target_position)


class QmixExternalValve(QmixValve):
    """
    An external valve, controlled by Qmix I/O-B.

    Parameters
    ----------
    index : int, or None
        Index of the DIO channel to istantiate. Takes precedence over the
        `name` parameter.

    name : str
        The name of the DIO channel to initialize. Will be ignored if `index` is
        not `None`.

    """
    def __init__(self, index=None, name=''):
        if index is None and name == '':
            raise ValueError('Please specify a valid DIO index or name')
        else:
            self.index = index
            self.name = name

        if self.index is not None:
            self._dio = QmixDigitalIO(index=self.index)
        else:
            self._dio = QmixDigitalIO(name=self.name)

        self.aspirate_pos = 1
        self.dispense_pos = 0

    @property
    def number_of_positions(self):
        return 2

    @property
    def position(self):
        """
        The current logical valve position.

        Each valve position is identified by a logical valve position identifier
        from 0 to number of valve positions - 1. This function returns the
        logical valve position identifier for the current valve position.

        Returns
        -------
        int
           >=0 current valve position index.

        """
        r = self._dio.is_output_on

        if r:
            return 1
        else:
            return 0

    @position.setter
    def position(self, position):
        self.switch_position(position)

    def switch_position(self, position=None):
        """
        Switch the valve to a certain logical valve position.

        Parameters
        ----------
        position : int, or None
            Switch the valve to the specified position.
            If `None` and a valve with two possible positions is connected,
            switch from the current position to the other one.

        """
        if position is None:
            if self.position == 0:
                target_position = 1
            else:
                target_position = 0
        else:
            target_position = position

        if ((target_position < 0) or
                (target_position >= self.number_of_positions)):
            msg = ('Must specify position in the range [0, %i] for this valve.'
                   % (self.number_of_positions - 1))
            raise ValueError(msg)

        self._dio.write(target_position)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os


def CHK(return_code, *args):
    """
    Check if the return value of the invoked function returned an error.

    The Qmix DLL makes this pretty easy: All errors are indicated by
    negative return values.

    Parameters
    ----------
    return_code : int
        The code returned from a Qmix DLL function.
    args
        All arguments passed to the function ``funcname``.

    Returns
    -------
    return_code : int
        If no error occurred, the originally passed return value is
        returned.

    Raises
    ------
    RuntimeError
        If the DLL function returned an error code.

    """
    if return_code >= 0:
        return return_code
    else:
        from .error import QmixError  # We import here to avoid circularity with error.py
        e = QmixError(return_code)
        error_string = e.error_string
        msg = (error_string + ", Error number: " + str(e.error_number) +
               ", Error code: " + str(e.error_code))
        raise RuntimeError(msg)


def find_dll(dll_dir, dll_filename):
    """
    Try to find the specified DLL.

    If a DLL directory was given, try to find the DLL there.

    If no DLL directory was given, start looking for the DLL in the following
    order:

    - current working directory
    - global DLL search paths (this also respects the PATH
      environment variable)
    - default Qmix SDK installation directory

    Parameters
    ----------
    dll_dir : string or None
        The directory containing the DLLs. Pass `None` to look in the default locations.
    dll_filename : string
        The name of the DLL to find, including filename extension.

    Returns
    -------
    dll_path : string or None
        The path of the DLL, or `None` if the DLL could not be found.

    """
    if not dll_dir:
        # Check current directory.
        if os.path.exists(dll_filename):
            dll_path = dll_filename
            return dll_path

        # Check DLL search paths.
        import win32api
        try:
            win32api.LoadLibrary(dll_filename)
            dll_path = dll_filename
            return dll_path
        except Exception:
            pass

        # Check default installation directory.
        import appdirs
        dll_dir = appdirs.user_data_dir('QmixSDK', '')
        if os.path.exists(os.path.join(dll_dir, dll_filename)):
            dll_path = os.path.join(dll_dir, dll_filename)
            # Add DLL dir to the DLL search path, otherwise we won't be able to
            # load the DLL.
            os.environ['PATH'] = dll_dir + os.pathsep + os.environ['PATH']
            try:
                # Python >= 3.8
                os.add_dll_directory(dll_dir)
            except NameError:
                pass

            return dll_path

        return None
    else:
        if os.path.exists(os.path.join(dll_dir, dll_filename)):
            dll_path = os.path.join(dll_dir, dll_filename)
            # Add DLL dir to the DLL search path, otherwise we won't be able to
            # load the DLL.
            os.environ['PATH'] = dll_dir + os.pathsep + os.environ['PATH']
            try:
                # Python >= 3.8
                os.add_dll_directory(dll_dir)
            except NameError:
                pass

            return dll_path
        else:
            return None




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import time
from cffi import FFI

if sys.version_info[0] < 3:
    # Python 2 compatibility; requires `future` package.
    from builtins import bytes

from . import config
from .tools import CHK, find_dll
from .headers import BUS_HEADER


class QmixBus(object):
    """
    Qmix bus interface.

    This interface establishes a connection with the labbCAN bus used for
    communication with all attached devices. Accordingly, has to be
    initialized before any hardware can be accessed.

    Parameters
    ----------

    auto_open : bool
        Whether to open the labbCAN bus automatically on object instantiation.

    auto_start : bool
        Whether to start the CAN bus communication automatically on object
        instantiation. Since the bus needs to be opened before communication
        can commence, setting `auto_start=True` will always open the bus,
        regardless of the `auto_open` parameter specified.

    """

    def __init__(self, auto_open=True, auto_start=True):
        dll_dir = config.read_config().get('qmix_dll_dir', None)
        dll_filename = 'labbCAN_Bus_API.dll'

        dll_path = find_dll(dll_dir=dll_dir, dll_filename=dll_filename)
        if dll_path is None:
            msg = 'Could not find the Qmix SDK DLL %s.' % dll_filename
            raise RuntimeError(msg)
        else:
            self.dll_path = dll_path

        self._ffi = FFI()
        self._ffi.cdef(BUS_HEADER)
        self._dll = self._ffi.dlopen(self.dll_path)

        config_dir = config.read_config().get('qmix_config_dir')
        if config_dir is not None:
            self.config_dir = config_dir
        else:
            msg = ('Please specify the Qmix configuration via '
                   'pyqmix.config.set_qmix_config() first.')
            raise RuntimeError(msg)

        self.auto_open = auto_open
        self.auto_start = auto_start

        self._p_config_dir = self._ffi.new(
            'char[]',
            bytes(self.config_dir, 'utf8'))

        self._p_plugin_search_path = self._ffi.NULL

        self.is_open = False
        self.is_started = False

        if self.auto_open:
            self.open()

        if self.auto_start:
            if not self.is_open:
                self.open()
            self.start()

    def __del__(self):
        self.stop()
        self.close()

    def _call(self, func_name, *args):
        func = getattr(self._dll, func_name)
        r = func(*args)
        return CHK(r)

    def open(self):
        """
        Initialize labbCAN bus.

        Initializes resources for a labbCAN bus instance, opens the bus and
        scans for connected devices.

        """
        self._call('LCB_Open',
                   self._p_config_dir,
                   self._p_plugin_search_path)
        time.sleep(1)
        self.is_open = True

    def close(self):
        """
        Close labbCAN bus.

        """
        self._call('LCB_Close')
        self.is_open = False

    def start(self):
        """
        Start bus network communication.

        Sets all connected devices operational and enables them.
        Connected devices can be accessed only after this method has been
        invoked.

        """
        if not self.is_open:
            msg = ('Bus needs to be opened before communication can start.'
                   'Call `QmixBus.open()` first.')
            raise RuntimeError(msg)

        self._call('LCB_Start')
        time.sleep(1)
        self.is_started = True

    def stop(self):
        """
        Stop bus network communication.

        Stops network communication and closes the labbCAN device.
        The method should be called before calling
        :func:`qmix.QmixBus.close``.

        """
        self._call('LCB_Stop')
        self.is_started = False
